package com.fossgalaxy.games.tbs.actions;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.parameters.ResourceType;
import com.fossgalaxy.games.tbs.order.MineOrder;
import com.fossgalaxy.games.tbs.order.Order;
import com.fossgalaxy.object.annotations.ObjectDef;
import org.codetome.hexameter.core.api.CubeCoordinate;


/**
 * A mine action executes an order to generate a resource. It can only be executed by an entity on a resource.
 *
 * This action can be automated - failing to provide any order to an entity that has this action will cause it to
 * execute this action at the end of the turn.
 */
public class MinesAction extends AbstractMineAction {

    private final int quantityPerTurn;

    @ObjectDef("Mine")
    public MinesAction(ResourceType resourceType, int quantityPerTurn) {
        super(resourceType);
        this.quantityPerTurn = quantityPerTurn;
    }

    @Override
    public Order generateOrder(CubeCoordinate cubeCoordinate, GameState gameState) {
        return new MineOrder(resourceType, quantityPerTurn);
    }
}
