package com.fossgalaxy.games.tbs.actions;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.entity.Entity;
import com.fossgalaxy.games.tbs.order.SellOrder;
import com.fossgalaxy.games.tbs.order.Order;
import com.fossgalaxy.object.annotations.ObjectDef;
import org.codetome.hexameter.core.api.CubeCoordinate;

import java.awt.*;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Locale;

public class SellAction extends AbstractAction {

    private final double regen;
    private final String toString;

    @ObjectDef("Sell")
    public SellAction(double regen) {
        this.regen = regen;
        DecimalFormat df = new DecimalFormat("#.###", new DecimalFormatSymbols(Locale.US));
        this.toString = String.format("Sell for %s of value", df.format(regen));
    }

    @Override
    public boolean isPossible(Entity entity, GameState s, CubeCoordinate co) {
        return entity.getPos().equals(co);
    }

    @Override
    public Order generateOrder(CubeCoordinate co, GameState s) {
        return new SellOrder(regen);
    }

    @Override
    public Color getHintColour() {
        return AbstractAction.HINT_ATTACK_COLOUR;
    }

    @Override
    public Color getBorderColour() {
        return Color.RED;
    }

    @Override
    public String toString(){
        return toString;
    }
}
