package com.fossgalaxy.games.tbs.ai.rules;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.entity.Entity;
import com.fossgalaxy.games.tbs.entity.Resource;
import com.fossgalaxy.games.tbs.order.Order;
import com.fossgalaxy.games.tbs.parameters.ResourceType;
import com.fossgalaxy.games.tbs.ui.GameAction;
import com.fossgalaxy.object.annotations.ObjectDef;
import org.codetome.hexameter.core.api.CubeCoordinate;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Created by pwillic on 18/05/2018.
 */
public class UseActionOnResource extends PerEntityRule {

    private final GameAction action;
    private final List<ResourceType> targets;

    @ObjectDef("UseActionOnResource")
    public UseActionOnResource(GameAction action, ResourceType... targets) {
        this.action = action;
        this.targets = Arrays.asList(targets);
    }

    @Override
    public boolean isForEntity(GameState state, Entity entity) {
        return entity.getType().getAvailableActions().contains(action);
    }

    @Override
    public Order generateOrder(GameState gameState, Entity entity) {
        CubeCoordinate ourPos = entity.getPos();
        Collection<Resource> resources = gameState.getResources().stream()
                .filter(x -> targets.contains(x.getType()))
                .collect(Collectors.toList());

        if(resources.isEmpty()) return null; // Unlikely but still
        Resource closestResource = null;
        double closestDist = Double.MAX_VALUE;
        for(Resource resource : resources){
            CubeCoordinate pos = resource.getLocation();
            if(!action.isPossible(entity, gameState, pos)) continue;
            double distance = gameState.getDistance(ourPos, pos);
            if(distance == 0)continue;
            if(distance < closestDist){
                closestDist = distance;
                closestResource = resource;
            }
        }
        if(closestResource == null) return null;
        return action.generateOrder(closestResource.getLocation(), gameState);
    }
}
